'''
    Project metadata is specified here.

    This module *should not* import anything from the project or third-party
    modules, to avoid dependencies in setup.py or circular import issues.
'''
from time import localtime as _localtime


pkgname         = 'tconf'
__version__     = version = '0.59a2'
__author__      = authors = ', '.join([
                                'Mike Miller',
                                #~ 'and contributors',
                            ])
copyright       = '© 2019-%s' % _localtime().tm_year
description     = "It's turtles all the way down…"
license         = 'LGPLv3'
keywords        = 'configuration cascade cascading ini conf yaml xml environment'

# online repo information
repo_account    = 'mixmastamyk'
repo_name       = pkgname
repo_provider   = 'github.com'
#~ repo_provider   = 'bitbucket.org'
repo_url        = 'https://%s/%s/%s' % (repo_provider, repo_account, repo_name)
project_urls    = {'Repository': repo_url, 'Issues': repo_url + '/issues'}
email           = 'mixmastamyk@%s' % repo_provider

trove_classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'Topic :: System :: Systems Administration',
    'Topic :: Utilities',
]


class defaults:
    pass



'''
    | tconf - TurtleConfig Adapters
    | © 2020, Mike Miller - Released under the LGPL, version 3+.

    Adapt various config sources to a single style of interface.
'''
import os
import logging


log = logging.getLogger(__name__)


class _AttributeDict(dict):
    ''' Access dict items through attributes '''
    def __getattr__(self, attr):
        return self.get(attr)

    __getitem__ = __getattr__


class _Adapter:
    ''' Abstract Base. '''
    def __repr__(self):
        return f'{self.__class__.__name__}({self._source!r})'


class ArgParserAdapter(_Adapter):
    ''' Wraps a TurtleArgumentParser. '''
    def __init__(self, source):
        # parse just once, use Namespace afterward
        self._source = source.parse_args()

    def __getattr__(self, attr_name, **kwargs):
        flattened_name = attr_name.replace('.', '_')
        log.debug('%s.get(%r)', self.__class__.__name__, flattened_name)
        return getattr(self._source, flattened_name, None)

    def __repr__(self):
        return f'{self.__class__.__name__}( {self._source!r} )'


class ConfigParserAdapter(_Adapter):
    ''' Loads values from .ini format files via ConfigParser.
        Note: this supports only one or two levels of hierarchy.
    '''
    def __init__(self, file_path, interpolation=False,
                 default_section=None, **kwargs):
        # defer to avoid loading when not needed:
        from configparser import ConfigParser
        self._copa = ConfigParser(interpolation=interpolation)
        self._copa.read(file_path)
        self._def_sect = default_section
        self._source = file_path

    def __getattr__(self, attr_name):
        log.debug('%s.get(%r)', self.__class__.__name__, attr_name)
        import configparser as cp
        value = None
        section, _, name = attr_name.partition('.')
        if name:  # 2 or more levels
            try:
                value = self._copa.get(section, name)
                log.debug('  found .ini option: [%s] %s = %r',
                          section, name, value)
            except (cp.NoSectionError, cp.NoOptionError) as err:
                log.debug('  ' + str(err))
        else:  # there was only one level this time
            if self._copa.has_section(section):
                value = _AttributeDict(self._copa[section])
            elif self._copa.has_section(self._def_sect):
                # fallback to default
                log.debug('  falling back to [%s]%s', self._def_sect, attr_name)
                value = self._copa[self._def_sect].get(attr_name)
                log.debug('  value: %r %r', value, type(value))
            else:
                pass

        return value


class EnvAdapter(_Adapter):
    ''' Finds values set in the system environment.

        Limitation: Can't find hierarchical values unless used with dict
                    notation.
    '''
    def __init__(self, prefix, env=os.environ, **kwargs):
        self._prefix = prefix
        self._source = env
        self._key = ''

    def __getattr__(self, attr_name):
        key = '.'.join([self._prefix,
                        attr_name.upper()])
        log.debug('%s.get(%r)', self.__class__.__name__, key)

        value = self._source.get(key)
        if value is not None:
            log.info('  found environment variable: %r %r', key, value)
        else:
            log.debug('  not found in environment.')

        return value

    def __repr__(self):
        return f'{self.__class__.__name__}(os.environ)'


class JSONAdapter(_Adapter):
    ''' Loads values from JSON format files. '''
    def __init__(self, file_path, **kwargs):
        from json import load  # defer to avoid loading when not needed
        with open(file_path) as f:
            self._data = load(f, object_hook=_AttributeDict)
        self._source = file_path

    def __getattr__(self, attr_name):
        log.debug('%s.get(%r)', self.__class__.__name__, attr_name)
        value = self._data

        for segment in attr_name.split('.'):  # crawl down
            value = getattr(value, segment, None)
            log.debug('  segment: %r, value: %r', segment, value)
            if value is None:
                break

        return value


class ObjectAdapter(_Adapter):
    ''' Load values from objects. '''
    def __init__(self, source):
        self._source = source

    def __getattr__(self, attr_name, **kwargs):
        log.debug('ObjectAdapter.get attr_name: %r', attr_name)
        value = self._source
        for segment in attr_name.split('.'):  # crawl down
            value = getattr(value, segment, None)
            log.debug('  segment: %r, val: %r', segment, value)
            if value is None:
                break

        if value is not None:
            if isinstance(value, type):  # is class type
                log.debug('  class found: %r', value)
                value = self.__class__(value)
            log.debug('  found: %s=%r', attr_name, value)
        else:
            log.debug('  not found in object.')

        return value

    def __repr__(self):
        return f'{self.__class__.__name__}({self._source.__name__})'


class SYAMLAdapter(_Adapter):
    ''' Loads values from YAML format files. '''
    def __init__(self, file_path, **kwargs):
        import strictyaml  # defer to avoid loading when not needed
        with open(file_path) as f:
            self._data = strictyaml.load(f.read())
        self._yaml_type = strictyaml.YAML
        self._source = file_path

    def __getattr__(self, attr_name):
        value = self._data  # start here
        log.debug('%s.get(%r)', self.__class__.__name__, attr_name)

        for segment in attr_name.split('.'):  # crawl down
            value = value.get(segment)
            log.debug('  segment: %r, value: %s', segment, value)
            if value is None:
                break
            elif isinstance(value, self._yaml_type):
                try:
                    value = _AttributeDict(value.data)
                except ValueError:
                    value = value.data
                    log.info('  found yaml option: %s: %r', segment, value)

        return value


class XMLAdapter(_Adapter):
    ''' Loads values from XML format files and directs access.

        Note: skips root element to provide parity with other source types.
    '''
    def __init__(self, file_path, attr_prefix='_', **kwargs):
        import xmltodict  # defer to avoid loading when not needed
        with open(file_path) as f:
            data = xmltodict.parse(
                f.read(), xml_attribs=True, attr_prefix=attr_prefix,
                disable_entities=False,
            )
        if len(data) == 1:  # skip root
            data = tuple(data.values())[0]
        self._data = data
        self._source = file_path

    def __getattr__(self, attr_name):
        from collections import OrderedDict
        value = _AttributeDict(self._data)

        for segment in attr_name.split('.'):  # crawl down
            value = getattr(value, segment, None)
            log.debug('segment: %r, value: %r', segment, value)
            if value is None:
                break
            elif isinstance(value, OrderedDict):  # no object_hook
                value = _AttributeDict(value)
                # needs to crawl for more OrderedDicts
        return value


file_adapter_map = {
    '.ini': ConfigParserAdapter,
    '.json': JSONAdapter,
    '.xml': XMLAdapter,
    '.yml': SYAMLAdapter,
    '.yaml': SYAMLAdapter,
}



import sys
from os.path import dirname, join
from setuptools import setup
from tconf import meta


if sys.version_info < (3, 6):
    raise NotImplementedError('Sorry, only Python 3.6 and above is supported.')

# https://www.python.org/dev/peps/pep-0508/#environment-markers
install_requires = (
    'appdirs',
    'typeguard',
)
tests_require = ()  # ('pyflakes', 'readme_renderer'),
extras_require = dict(
    yaml=('strictyaml',),
    xml=('xmltodict',),
)

def slurp(filename):
    try:
        with open(join(dirname(__file__), filename), encoding='utf8') as infile:
            return infile.read()
    except FileNotFoundError:
        pass  # needed at upload time, not install time


setup(
    name                = meta.pkgname,
    description         = meta.description,
    author_email        = meta.email,
    author              = meta.authors,
    keywords            = meta.keywords,
    license             = meta.license,
    long_description    = slurp('readme.rst'),
    url                 = meta.repo_url,
    packages            = (meta.pkgname,),
    project_urls        = meta.project_urls,
    version             = meta.version,

    extras_require      = extras_require,
    install_requires    = install_requires,
    python_requires     = '>=3.6',
    setup_requires      = install_requires,
    tests_require       = tests_require,

    classifiers         = meta.trove_classifiers,
)



'''
    | tconf - TurtleConfig - It's turtles all the way down…
    | © 2020, Mike Miller - Released under the LGPL, version 3+.
'''
import logging
import os
from argparse import ArgumentParser
from ast import literal_eval
from collections.abc import Sequence
from os.path import abspath, dirname, exists
from types import ModuleType
from typing import get_type_hints, _GenericAlias, _SpecialForm

from typeguard import check_type

from . import adapters, meta
from .adapters import file_adapter_map


log = logging.getLogger(__name__)


class DefaultsMissingError(RuntimeError):
    ''' Defaults not found, one must be passed in the source argument sequence. '''


class TurtleConfig:
    ''' A configuration object that retrieves parameters from a number of
        sources, in the order given.
    '''
    _env_prefix = 'PY'

    def __init__(self, app_name, sources,
                 ensure_paths=False,
                 env_prefix=None,
                 ini_default_section='main',
                 ini_interpolation=None,
                 vendor_name=None,
                ):
        log.debug('🐢 TurtleConfig, version: %r', meta.version)
        if not isinstance(sources, Sequence):
            raise ValueError('A sequence of sources is required.')

        self._app_name = app_name
        self._ini_default_section = ini_default_section
        self._ini_interpolation = ini_interpolation
        self._types_cache = {}
        self._values_cache = {}
        self._vendor_name = vendor_name
        if env_prefix:
            self._env_prefix = env_prefix

        # wrap sources with Adapters
        self._sources = []
        for source in sources:
            if isinstance(source, str):
                source = self._handle_path(source, ensure_paths)
            wrapped = self._adapt_source(source)
            if wrapped is not None:
                self._sources.append(wrapped)
            log.debug('  source: %r', wrapped)

        # find the defaults object, likely bringing up the rear:
        for obj in reversed(self._sources):
            if isinstance(obj, adapters.ObjectAdapter):
                self._types_cache = { p[0]: p[2]  # dump value
                                      for p in _list_object_props(obj._source) }
                break  # -en Sie
        else: # no break, aka not found
            raise DefaultsMissingError(DefaultsMissingError.__doc__)

    def __getattr__(self, attr_name):
        ''' Attribute-style interface: cfg.foo.bar.baz.

            Only called when self.attr doesn't exist.
        '''
        if attr_name in self._values_cache:
            return self._values_cache[attr_name]
        value = None
        log.debug('🐢.get(%r)', attr_name)

        # find value
        for source in self._sources:  # find the value
            value = getattr(source, attr_name)
            if value is not None:
                break  # found something
        else:  # not broken, not found
            raise AttributeError('%r not found.' % attr_name)

        # fix attr_name if used with default section in ConfigParser
        # This is very complicated, would like to remove this:
        if ('.' not in attr_name
            and isinstance(source, adapters.ConfigParserAdapter)
            and source._def_sect
            and not isinstance(value, adapters._AttributeDict)):
                attr_name = source._def_sect + '.' + attr_name

        # potentially convert then type check value
        dest_type = self._types_cache.get(attr_name)
        if isinstance(value, str) and dest_type is not str:
            # strings may or may not need type coercion
            value = self._coerce_string(attr_name, value, dest_type)

        elif isinstance(value, adapters.ObjectAdapter):
            return value  # needed for attr iface :-/

        if isinstance(value, adapters._AttributeDict):  # very inefficient,
            # checks/converts whole section to support the attr iface :-/
            for key, val in value.items():  # every one !
                key_name = attr_name + '.' + key
                dest_type = self._types_cache.get(key_name)
                if isinstance(val, str) and dest_type is not str:
                    value[key] = self._coerce_string(key_name, val, dest_type)
                check_type(key_name, value[key], dest_type)
        else: # everything else
            check_type(attr_name, value, dest_type)

        self._values_cache[attr_name] = value
        return value

    def __getitem__(self, attr_path):
        ''' Dictionary style interface: cfg['foo.bar.baz'].

            TODO: make this the main implementation.
        '''
        try:
            return self.__getattr__(attr_path)
        except AttributeError as err:
            raise KeyError(str(err)) from err

    def _adapt_source(self, source):
        ''' Wrap a source with an Adapter class. '''
        if isinstance(source, adapters._Adapter):   # not again!
            pass
        elif isinstance(source, str):               # a path
            pth = source
            source = None
            if os.access(pth, os.R_OK):  # avoid non-existent | unreadable file
                _, ext = os.path.splitext(pth)
                AdapterClass = file_adapter_map.get(ext.casefold())
                if AdapterClass:
                    source = AdapterClass(pth,
                        interpolation=self._ini_interpolation,
                        default_section=self._ini_default_section,
                    )
            else:
                log.warn('file %r unreadable, skipping.', pth)

        elif source in (os.environ, os.environb):
            prefix = self._env_prefix + '_' + self._app_name.upper()
            source = adapters.EnvAdapter(prefix, env=source)

        elif isinstance(source, TurtleArgumentParser):
            source = adapters.ArgParserAdapter(source)

        elif isinstance(source, (type, ModuleType)):  # is class or module
            source = adapters.ObjectAdapter(source)

        else:
            raise NotImplementedError('Source %r not recognized.' % source)

        return source

    def _coerce_string(self, attr_name, value, type_):
        ''' Convert a string value to the expected type. '''
        if isinstance(type_, dict):  # retrieve from argparse kwargs
            type_ = type_.get('type')

        if type_ is not None:
            if type_ is bool:
                converted = value.casefold() in ('true', '1')

            # compound types will need a harder look:
            elif (type_ in (list, dict, set, tuple) or
                  isinstance(type_, (_GenericAlias, _SpecialForm)) # __origin__
                ):  # a safer, limited eval, raises (ValueError, SyntaxError)
                converted = literal_eval(value)

            elif type_ is type(None):  # type_ is NoneType  :-/
                if value is None or value.casefold() in ('null', 'none'):
                    converted = None

            else:  # simple types: int, float, …
                converted = type_(value)

            log.debug('coercing value from: %r to %r', value, converted)
            value = converted

        return value

    def _handle_path(self, path_str, ensure_paths=False):
        ''' Render an absolute path with folders from appdirs,
            and optionally ensure file exists.
        '''
        if path_str.startswith('{'):  # try to render template
            from appdirs import site_config_dir, user_config_dir
            sd = site_config_dir(self._app_name, self._vendor_name)
            ud = user_config_dir(self._app_name, self._vendor_name)
            path_str = path_str.format(site_config_dir=sd, user_config_dir=ud)

        if ensure_paths:
            folder_exists = None  # these are separate, to create relative file
            folder = dirname(path_str)
            # avoid attempt to create own folder:
            if folder in ('', '.'):  # current dir
                folder_exists = True
            else:
                log.debug('ensuring %r', path_str)
                try:
                    os.makedirs(folder, exist_ok=True)
                    folder_exists = True
                except OSError as err:
                    log.warn('unable to create path: %s', str(err))

            if folder_exists:
                try:
                    if not exists(path_str):  # race
                        open(path_str, 'a').close()
                except OSError as err:
                    log.warn('unable to create file: %s', str(err))

        if not path_str.startswith('/'):  # relative?
            if len(path_str) > 2 and path_str[1] == ':':
                pass  # Winders, nope
            else:
                path_str = abspath(path_str)
        return path_str

    def add_turtle_source(self, source):
        ''' After the fact. '''
        if not isinstance(source, adapters._Adapter):
            source = self._adapt_source(source)
        self._sources = self._sources + (source,)

    def clear_turtle_cache(self):
        ''' Use to update variables after a config update, or free memory. '''
        self._values_cache.clear()


class TurtleArgumentParser(ArgumentParser):
    ''' An ArgumentParser that is automatically populated by TurtleConfig.

        Arguments:
            app_defaults:  a (class, module, object) containing data.
            help_templ: a string such as: '🐢 {description} ({type_str})'
    '''
    def __init__(self, app_defaults, *args, help_templ=None, **kwargs):

        super().__init__(*args, **kwargs)
        if not help_templ:
            help_templ = '🐢 {description} ({type_str})'

        # look over default object and configure argument details:
        for key, value, annotation in _list_object_props(app_defaults, mod_name=1):

            log.debug('TurtleArgumentParser arg: %r', (key, value, annotation))
            params = {}
            if type(annotation) is dict:
                description = annotation.pop('desc', '')
                type_ = annotation.pop('type', value.__class__)
                params = annotation  # pass rest to parser.add_argument()
            else:
                type_, description = value.__class__, ''
            type_str = type_.__name__ if hasattr(type_, '__name__') else str(type_)

            if value is False:
                params['action'] = 'store_true'
                type_str = 'False, sets True'
            elif value is True:
                params['action'] = 'store_false'
                type_str = 'True, sets False'
            else:
                params['metavar'] = type_str[0].upper()
                params['type'] = type_

            if type(annotation) is dict and 'choices' in annotation:
                type_str += ': ' + str(annotation['choices'])[1:-1]

            # specific help value overrides
            if 'help' not in params:
                params['help'] = help_templ.format(
                    description=description,
                    type_str=type_str,
                )
            # build argument
            self.add_argument(
                (self.prefix_chars[0]*2) + key,
                default=None, # don't want to stop here, continue with None
                **params,
            )


def _list_object_props(container, prefix='', mod_name=False):
    ''' Inspect object property annotations and types, return a list.

        Arguments:
            container:  the object to inspect
            prefix:     keeps track of where we are in tree
            mod_name:   use argparse style names instead of obect notation.
    '''
    annos = get_type_hints(container)
    arg_list = []
    if prefix:
        if mod_name:
            prefix = prefix + '-'
        else:
            prefix = prefix + '.'

    for key, value in vars(container).items():
        if key.startswith('_'):
            continue

        annotation = annos.get(key)
        if mod_name:
            arg_name = prefix + key.replace('_', '-')
        else:
            arg_name = prefix + key

        if isinstance(value, type):  # class, follow container
            arg_list.extend(_list_object_props(value, prefix=arg_name))
        else:
            if annotation:
                arg_list.append( (arg_name, value, annotation) )
            else:  # None, fall back to type of value
                arg_list.append( (arg_name, value, type(value)) )

    return arg_list
